# SPDX-FileCopyrightText: 2025-present Daniel Skowroński <daniel@skowron.ski>
# base on https://github.com/abcminiuser/python-elgato-streamdeck/blob/master/src/example_neo.py by abcminiuser
#
# SPDX-License-Identifier: MIT
from PIL import Image, ImageDraw, ImageFont
from StreamDeck.DeviceManager import DeviceManager
from StreamDeck.ImageHelpers import PILHelper

from threading import Thread, Lock
import sched
import time
import datetime
import sys
import signal
from functools import partial
import traceback

import logging

from rpideck.cli.RPiDeckConfig import RPiDeckConfig
from rpideck.cli.DDC import DDCLinux
from rpideck.cli.AVR import AVR


class RPiDeck:
    def __init__(self, config_path="~/.config/rpideck", logger_name=__name__):
        self.logger = logging.getLogger(logger_name)
        self.config = RPiDeckConfig(config_path, logger_name)
        self.display_lock = Lock()
        self.deck = None
        self.page = 0
        self.ddcController = DDCLinux()
        self.avrController = AVR(self.config.avr["ip"])
        sys.excepthook = self.exceptionHandler
        self.clock_scheduler = sched.scheduler(time.time, time.sleep)

    def openDeck(self, matchSerial=""):
        streamdecks = DeviceManager().enumerate()
        signal.signal(signal.SIGINT, partial(self.sigintHandler))
        for deck in streamdecks:
            deck.open()
            self.logger.debug(deck.get_serial_number())
            if matchSerial in deck.get_serial_number():
                self.deck = deck
            deck.close()
        if self.deck is None:
            raise Exception(f"didn't find working deck matching {matchSerial}")
        self.deck.open()
        self.logger.info(
            "Opened '{}' device (serial number: '{}', fw: '{}')".format(
                self.deck.deck_type(),
                self.deck.get_serial_number(),
                self.deck.get_firmware_version(),
            )
        )  # FIXME

    def initializeDeck(self):
        self.deck.reset()
        self.deck.set_poll_frequency(100)
        self.deck.set_brightness(self.config.deck["brightness"])
        self.loadPage(0)
        self.deck.set_key_callback(self.keyChangeCallback)
        self.updateScreenText("RPiDeck started")
        self.clock_scheduler.enter(0.25, 1, self.updateClock)
        self.clock_scheduler.run()

    def loadPage(self, page=0):
        self.page = page
        for key in range(self.deck.key_count()):
            self.updateKeyImage(key, page, False)

    def updateKeyImage(self, position, isPressedDown, page, background="black"):
        try:
            keyInfo = self.config.getKeyInfo(position, self.page, isPressedDown)
            keyImage = self.renderKeyImage(
                keyInfo["icon"], keyInfo["font"], keyInfo["label"], background
            )
            with self.deck:
                self.deck.set_key_image(position, keyImage)
        except Exception:
            with self.deck:
                self.deck.set_key_color(position, 0, 0, 0)

    def renderKeyImage(
        self, icon_filename, font_filename, label_text, background="black"
    ):
        icon = Image.open(icon_filename)
        image = PILHelper.create_scaled_key_image(
            self.deck, icon, margins=[0, 0, 20, 0], background=background
        )
        draw = ImageDraw.Draw(image)
        font = ImageFont.truetype(font_filename, 14)
        draw.text(
            (image.width / 2, image.height - 5),
            text=label_text,
            font=font,
            anchor="ms",
            fill="white",
        )
        return PILHelper.to_native_key_format(self.deck, image)

    def updateScreenText(self, text):
        screenImage = self.renderScreenImage(text, self.config.getFont())
        with self.deck:
            self.deck.set_screen_image(screenImage)

    def renderScreenImage(self, text, font_filename):
        image = PILHelper.create_screen_image(self.deck)
        draw = ImageDraw.Draw(image)
        font = ImageFont.truetype(font_filename, 22)
        draw.text(
            (image.width / 2, image.height - 25),
            text=text,
            font=font,
            anchor="ms",
            fill="white",
        )
        return PILHelper.to_native_screen_format(self.deck, image)

    def executeKeyAction(self, position, page):
        pass
        # TODO: implement logic separation from keyChangeCallback

    def keyChangeCallback(self, deck, position, isPressedDown):
        if position >= self.config.BUTTONS:
            if isPressedDown:
                delta = 0
                if position == self.config.BUTTONS:
                    self.logger.info(f"Key {position} pressed => calling page:previous")
                    delta = -1
                if position == self.config.BUTTONS + 1:
                    self.logger.info(f"Key {position} pressed => calling page:next")
                    delta = +1
                self.page = (self.page + delta) % self.getPageCount()
                self.loadPage(self.page)
            else:
                self.logger.info(f"Key {position} pressed up")
        else:
            keyInfo = self.config.getKeyInfo(position, self.page, isPressedDown)
            actionInfo = f"down => calling {keyInfo['name']}" if isPressedDown else "up"
            self.logger.info(f"Key {position} on page {self.page} pressed {actionInfo}")

            if isPressedDown:
                with self.deck:
                    self.updateKeyImage(
                        position,
                        isPressedDown,
                        self.page,
                        background=self.config.deck["highlightColour"],
                    )
                    with self.display_lock:
                        for step in keyInfo["steps"]:
                            self.updateScreenText(step["text"])
                            params = step["parameters"]
                            if step["type"] == "ddc":
                                self.ddcController.setVCP(
                                    params["vcp"],
                                    params["value"],
                                    params.get("force", False),
                                )
                            elif step["type"] == "eiscp":
                                self.avrController.cmd(params["cmd"], params["value"])
                        self.updateScreenText("callback finished")
                        self.updateKeyImage(position, isPressedDown, self.page)

    def getPageCount(self):
        return len(self.config.deck["pages"])

    def updateClock(self):
        now = datetime.datetime.now()
        if self.getPageCount() == 1:
            text = "{:04d}-{:02d}-{:02d} {:02d}:{:02d}:{:02d}".format(
                now.year, now.month, now.day, now.hour, now.minute, now.second
            )
        else:
            text = "page {:01d}/{:01d} | {:02d}:{:02d}:{:02d}".format(
                self.page + 1, self.getPageCount(), now.hour, now.minute, now.second
            )
        with self.display_lock:
            self.updateScreenText(text)
        self.clock_scheduler.enter(0.25, 1, self.updateClock)

    def exceptionHandler(self, exctype, value, tb):
        self.logger.exception("".join(traceback.format_exception(exctype, value, tb)))
        self.logger.exception("Uncaught exception: {0}".format(str(value)))

    def sigintHandler(self, sig, frame):
        self.logger.warning("caught SIGINT")
        with self.deck:
            self.deck.reset()
            self.deck.close()
        sys.exit(0)



# SPDX-FileCopyrightText: 2025-present Daniel Skowroński <daniel@skowron.ski>
# base on https://github.com/abcminiuser/python-elgato-streamdeck/blob/master/src/example_neo.py by abcminiuser
#
# SPDX-License-Identifier: MIT
import eiscp
import logging


class AVR:
    def __init__(self, ip, logger_name=__name__):
        self.logger = logging.getLogger(logger_name)
        self.ip = ip

    def cmd(self, cmd, value):
        expanded_cmd=f"{cmd} {value}"
        self.logger.info(f"eiscp cmd: {expanded_cmd}")
        with eiscp.eISCP(self.ip) as receiver:
            receiver.command(expanded_cmd)



# SPDX-FileCopyrightText: 2025-present Daniel Skowroński <daniel@skowron.ski>
# base on https://github.com/abcminiuser/python-elgato-streamdeck/blob/master/src/example_neo.py by abcminiuser
#
# SPDX-License-Identifier: MIT
from schema import Schema, And, Or, Use, Optional, SchemaError
import yaml
import io
import os
import logging


class RPiDeckConfig:
    BUTTONS = 8

    def _build_schema(self):
        PARAMETERS_DDC = Schema(
            {
                "vcp": lambda n: 0x00 <= n <= 0xFF,
                "value": lambda n: 0x00 <= n <= 0xFFFF,
                Optional("force"): bool,
            }
        )
        PARAMETERS_EISCP = Schema(
            {
                "cmd": str,
                "value": str,
            }
        )

        def validate_step(step):
            base_schema = Schema(
                {
                    "text": str,
                    "type": And(str, lambda t: t in ["ddc", "eiscp"]),
                    "parameters": dict,
                }
            )
            base_schema.validate(step)

            if step["type"] == "ddc":
                PARAMETERS_DDC.validate(step["parameters"])
            elif step["type"] == "eiscp":
                PARAMETERS_EISCP.validate(step["parameters"])
            return step

        KEY_SCHEMA = Schema(
            {
                "position": lambda n: 0 <= n <= self.BUTTONS,
                "icon": str,
                "label": str,
                "steps": [validate_step],
            }
        )
        PAGE_SCHEMA = Schema({"keys": {str: KEY_SCHEMA}})
        return Schema(
            {
                "ddc": object, # TODO: implement this
                "avr": {
                    "ip": str,
                },
                "logging": {
                    "level": And(str, lambda t: t in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
                },
                "deck": {
                    "brightness": lambda n: 0 <= n <= 100,
                    "matchSerial": str,
                    "font": str,
                    "highlightColour": str,
                    "pages": [PAGE_SCHEMA],
                },
            }
        )

    def __init__(self, path, logger_name=__name__):
        self.logger = logging.getLogger(logger_name)
        cfg_path = os.path.join(os.path.expanduser(path), "rpideck.yml")
        self.assets_path = os.path.join(os.path.expanduser(path), "assets")

        with open(cfg_path) as stream:
            self.raw_config = yaml.safe_load(stream)

        self.schema = self._build_schema()
        validated = self.schema.validate(self.raw_config)
        self.ddc = validated["ddc"]
        self.avr = validated["avr"]
        self.deck = validated["deck"]
        self.logging = validated["logging"]

    def getKeyInfo(self, position, page=0, isPresssedDown=False):
        key_name = None
        key_cfg = None
        page = self.deck["pages"][page]
        if page is None:
            raise Exception(f"no such page {page}")
        for k, v in page["keys"].items():
            if v["position"] == position:
                key_name = k
                key_cfg = v
                break
        if key_cfg is None:
            raise Exception(f"no such key position {position} on page {page}")

        return {
            "name": key_name,
            "icon": os.path.join(self.assets_path, key_cfg["icon"]),
            "font": self.getFont(),
            "label": key_cfg["label"],
            "steps": key_cfg["steps"],
        }

    def getFont(self):
        return os.path.join(self.assets_path, self.deck["font"])



# SPDX-FileCopyrightText: 2025-present Daniel Skowroński <daniel@skowron.ski>
# base on https://github.com/abcminiuser/python-elgato-streamdeck/blob/master/src/example_neo.py by abcminiuser
#
# SPDX-License-Identifier: MIT
import subprocess
import logging


class DDCLinux:
    def __init__(self, logger_name=__name__):
        self.logger = logging.getLogger(logger_name)

    def readVCP(self, vcp):
        current = subprocess.run(
            ["/usr/bin/ddcutil", "getvcp", "--brief", "0x{:02X}".format(vcp)],
            capture_output=True,
            text=True,
        )
        self.logger.info(current)
        if current.returncode != 0:
            self.logger.warning(f"getvcp for {vcp} exited with {current.returncode}")
            return None
        current_out = str(current.stdout).split("\n")[0].split(" ")
        current_value = int(f"0{current_out[3]}", 16)  # FIXME: multibyte values
        return current_value

    def setVCP(self, vcp, new_value, force=False):
        if not force:
            current_value = self.readVCP(vcp)
            if current_value is None:
                return
            if current_value == new_value:
                return
        action = subprocess.run(
            [
                "/usr/bin/ddcutil",
                "setvcp",
                "--sleep-multiplier",
                "10",
                "--brief",
                "0x{:02X}".format(vcp),
                "0x{:02X}".format(new_value),
            ],
            capture_output=True,
            text=True,
        )
        self.logger.info(action)



# SPDX-FileCopyrightText: 2025-present Daniel Skowroński <daniel@skowron.ski>
# base on https://github.com/abcminiuser/python-elgato-streamdeck/blob/master/src/example_neo.py by abcminiuser
#
# SPDX-License-Identifier: MIT
import click
import threading
import logging
import coloredlogs
from StreamDeck.Transport.Transport import TransportError
from rpideck.__about__ import __version__
from rpideck.cli.RPiDeck import RPiDeck


@click.group(
    context_settings={"help_option_names": ["-h", "--help"]},
    invoke_without_command=True,
)
@click.version_option(version=__version__, prog_name="rpideck")
def rpideck():
    logger = logging.getLogger(__name__)
    logging.basicConfig(
        format="%(asctime)s %(levelname)-8s %(message)s",
        #level=logging.INFO,
        datefmt="%Y-%m-%d %H:%M:%S",
        #handlers=[logging.FileHandler("debug.log", mode="a"), logging.StreamHandler()],
    )

    rd = RPiDeck(
        "~/.config/rpideck"
    )  # FIXME: paramatrize this and split to config and assets path
    logger.setLevel(rd.config.logging["level"])
    coloredlogs.install(level=rd.config.logging["level"])

    rd.openDeck()
    rd.initializeDeck()
    for t in threading.enumerate():
        try:
            t.join()
        except (TransportError, RuntimeError):
            pass
